package ca.triumf.midas;

import java.net.*;
import java.io.*;
import java.util.*;

/**
 * The Experiment class creates a link to a Midas experiment and allows the user
 * to access all the parameters associated with the experiment. This class requires 
 * a socket based connection, and will not work using the Midas shared memory connection model.
 * <P>
 * The following simple example shows a connection made to a Midas server running on the local host.
 * Once the connection is made, an attempt is made to find a key and display its current value.
 * <P>
 *<B><code><pre>
public static void main(String args[]) {
  Experiment exp = new Experiment("localhost", "exp123", "JMidas");
  try {
    exp.connect();
    System.out.println("Client " + (exp.isClient(client, false) ? "" : "not ") + "connected");
    OdbKey key = new OdbKey(exp, "/Experiment");
    System.out.println(key.toString());
    exp.disconnect();
  } catch (Exception e) {
    System.out.println(e);
  }
}
 * </pre></code></B>
 *
 * @author Stefan Ritt - PSI
 * @author David Morris - TRIUMF
 */
public class Experiment extends java.lang.Object implements Serializable { //ActionListener, 
	/** The standard Midas TCP port */
  public static final int MIDAS_PORT = 1175;
	
	/** The version of Midas that this class was built for */
  public static final String MIDAS_VERSION = "1.9.1";

	/** The Midas error message flag */
  public static final int MERROR = (1<<0);
	/** The Midas information message flag */
  public static final int MINFO  = (1<<1);
	/** The Midas debug message flag */
  public static final int MDEBUG = (1<<2);
	/** The Midas user message flag */
  public static final int MUSER  = (1<<3);
	/** The Midas log message flag */
  public static final int MLOG   = (1<<4);
  public static final int MTALK  = (1<<5);
  public static final int MCALL  = (1<<6);
	/** The Midas all messages flag */
  public static final int MALL   = 0xFF;

	/** Run state is stopped. */
	public static final int STATE_STOPPED = 1;
	/** Run state is paused. */
	public static final int STATE_PAUSED  = 2;
	/** Run state is running. */
	public static final int STATE_RUNNING = 3;

	/** The Midas status code for successful communications operation */
	public static final String CM_SUCCESS                  = "1";
	/** The Midas status code for a set error */
	public static final String CM_SET_ERROR                = "102";
	/** The Midas status code for no client found */
	public static final String CM_NO_CLIENT                = "103";
	/** The Midas status code for a database access error */
	public static final String CM_DB_ERROR                 = "104";
	/** The Midas status code for an undefined or unknown experiment */
	public static final String CM_UNDEF_EXP                = "105";
	/** 
	 * The Midas status code for a software version mismatch. Results from mixing
	 * earlier and later versions of Midas servers and clients
	 */
	public static final String CM_VERSION_MISMATCH         = "106";
	/** The Midas status code for an attempt at accessing a shutdown server */
	public static final String CM_SHUTDOWN                 = "107";
	/** The Midas status code for an invalid password */
	public static final String CM_WRONG_PASSWORD           = "108";
	/** The Midas status code for an invalid environment */
	public static final String CM_UNDEF_ENVIRON            = "109";
	public static final String CM_DEFERRED_TRANSITION      = "110";
	public static final String CM_TRANSITION_IN_PROGRESS   = "111";
	
	/** The Midas status code for successful database operation */
	public static final String DB_SUCCESS                  = "1";

	
  private String host = "";
  private String experiment = "";
  private String client = "";
  private String password = "";
	private String odbName = "odb";
	
  private Socket sockSend;
	private Socket sockRecv;
	private Socket sockEvent;
	
  private PrintWriter pw;
  private BufferedOutputStream bw;
  private InputStream is;
  private InputStream recvIs;

  private static final int RPC_TIMEOUT = 10000;
  
  private int hDB = 0;
  private int hBufMsg, hBufEvent;
  
  private java.util.Timer  timerHotLink;
  private Vector vecHotLink = new Vector();
  
  private MessageListener messageListener = null;

	private OdbKey runState = null;
	
	/**
	 * A null constructor. The host, experiment and client values will need to be set before
	 * a connection can be made.
	 */
  public Experiment() {
  }
	
	/**
	 * Creates an Experiment object and defines the principle attributes.
	 *
	 * @param host the target host system. Can be <B>localhost</B>. If left blank, then localhost will be used.
	 * @param experiment the name of the experiment on the host.
	 * @param client the name of the calling client
	 */	
  public Experiment(String host, String experiment, String client) {
    this.host = host;
    this.experiment = experiment;
    this.client = client;
  }

	/**
	 * Sets or changes the name of the host after the Experiment is created. If left blank, then localhost will be used.
	 *
	 * @param host the name of the host running the Midas server. Can be <B>localhost</B>.
	 */	
  public void setHost(java.lang.String host) {
    this.host = host;
  }

	/**
	 * Returns the name of the host for this Experiment.
	 *
	 * @return the name of the host machine. If the host field is blank, 
	 * the experiment will use localhost, but report a blank host name.
	 */	
  public java.lang.String getHost() {
    return this.host;
  }

	/** 
	 * Sets or changes the name of the Midas experiment after creation of the Experiemnt object.
	 *
	 * @param experiment the name of the Midas experiment
	 */	
  public void setExperiment(String experiment) {
    this.experiment = experiment;
  }

	/**
	 * Returns the name of the Midas experiment.
	 *
	 * @return the name of the experiment
	 */	
  public java.lang.String getExperiment() {
    return this.experiment;
  }

	/**
	 * Sets or changes the name of the client Java program.
	 *
	 * @param client the name of the client program
	 */	
  public void setClient(String client) {
    this.client = client;
  }

	/**
	 * Returns the name of the client Java program.
	 *
	 * @return the name of the client program
	 */	
  public java.lang.String getClient() {
		return this.client;
  }

	/**
	 * Sets the name of the odb command used in RPC calls. This may be a system specific
	 * problem. Depending on the target Midas server, it may be necessary
	 * to fully qualify the pathname of the odbedit command.
	 *
	 * @param odb name of the odb command on the target server
	 *
	 * @return false if no errors
	 */
	public boolean setOdbName(String odb) {
		odbName = odb;
		return false;
	}
	
	/**
	 * Returns the name of the odb command used in RPC calls.
	 *
	 * @return the name of the odb command used in the RPC calls.
	 */
	public String getOdbName() {
		return odbName;
	}
	
	/**
	 * Returns the handle to the database.
	 *
	 * @return the handle of the database
	 */	
  public int getHdb() {
    return this.hDB;
  }

	/**
	 * Passes a password to the host for connections requiring one
	 *
	 * @param password a password for the experiment
	 */
	public void setPassword(String password) {
		this.password = password;
	}
	
	/**
	 * Connects to the host via sockets and establishes a persistent
	 * link to the Midas experiment database. The connection is made initially using the MIDAS_PORT
	 * port, then three general use port numbers are passed to the Midas server for permanent connection.
	 *
	 * If the Midas experiment requires
	 * a password, the value of the password parameter is sent. This implies that the 
	 * user has set a password with the setPassword() method.
	 *
	 * @throws UnknownExpException the experiment is unknown on the host system
	 * @throws IOException a error occurred creating the socket connection to the host
	 * @return true if bad password else false
	 */	
  public boolean connect() throws UnknownExpException, IOException {
		
    if (host.equals("")) {     // use local host as default
      host = java.net.InetAddress.getLocalHost().getHostName();
		}
    
    ServerSocket s1 = new ServerSocket(0);	// create server sockets
    ServerSocket s2 = new ServerSocket(0);	// let system pick ports
    ServerSocket s3 = new ServerSocket(0);

		Socket s = new Socket(host, MIDAS_PORT);	// socket to talk to Midas
    is = s.getInputStream();
//    pw = new PrintWriter(s.getOutputStream());
    bw = new BufferedOutputStream(s.getOutputStream());
    s.setSoTimeout(RPC_TIMEOUT);

		String reply = rpcCall("C " + 
													 Integer.toString(s1.getLocalPort()) + " " +
                   				 Integer.toString(s2.getLocalPort()) + " " +
                   				 Integer.toString(s3.getLocalPort()) + " " +
                   				 MIDAS_VERSION + " " + 
                   				 (experiment.length() > 0 ? experiment : "Default"));

    s.close();	// finished with client socket now
    
    if (reply.charAt(0) == '2') {	// first char in response denotes error
      throw new UnknownExpException("Experiment " + experiment);
		}
      
    // check version
    StringTokenizer st = new StringTokenizer(reply.substring(2), " .");
    StringTokenizer vst = new StringTokenizer(MIDAS_VERSION, ".");
    if (!st.nextToken().equals(vst.nextToken()) ||
        !st.nextToken().equals(vst.nextToken())) {
			System.out.println("Server MIDAS version (" + reply.substring(2).trim() +  
                         ") differs from local version (" + MIDAS_VERSION + ")");
			return true;
		}
		
    // accept callback from Midas on all three sockets
    sockSend  = s1.accept();
    sockRecv  = s2.accept();
    sockEvent = s3.accept();
    
    s1.close();
    s2.close();
    s3.close();

    // set timeouts
    sockSend.setSoTimeout(RPC_TIMEOUT); // for RPC reply
    sockRecv.setSoTimeout(1);           // for hotlink notification

		// define streams for communicating with Midas
//    pw = new PrintWriter(sockSend.getOutputStream());
    bw = new BufferedOutputStream(sockSend.getOutputStream());
    is = sockSend.getInputStream();
    recvIs = sockRecv.getInputStream();

		reply = rpcCall("128 " + client);
		
    // open experiment database
    reply = rpcCall("db_open_database&ODB&100000&" + client);
    hDB = Integer.parseInt(replyParam(reply, 1));

    // set client info in ODB. If password requested, send it
    reply = rpcCall("cm_set_client_info&" + hDB + "&" + InetAddress.getLocalHost().getHostName() + "&" + client + "&128&" + "");
		if (replyParam(reply, 0).equals(CM_WRONG_PASSWORD)) {
      String pwd = crypt(password, "mi");
      reply = rpcCall("cm_set_client_info&" + hDB + "&" + InetAddress.getLocalHost().getHostName() + "&" + client + "&128&" + pwd);
			if (replyParam(reply, 0).equals(CM_WRONG_PASSWORD)) {
				return true;
			}
		}
			
   
		// Send start notification  
		messageSend(MINFO, "Program " + client + " on host " +
								InetAddress.getLocalHost().getHostName() + " started");
		hotLink();
		return false;
  }
  
	/**
	 * Sends the passed message via the Send socket output stream. At this time, there
	 * is a problem with this method if the receiving socket blocks. The response does not time out
	 * and the method can hang!
	 *
	 * @param arg the string to send to the socket
	 * @throws IOException on any socket error
	 * @return a String with the reply from Midas
	 */	
  public synchronized String rpcCall(String arg) throws IOException {
    int n;
		byte s[];
    byte b[] = new byte[10000];
    String reply = "";
    
		s = arg.getBytes();
		
    // send arguments
//    pw.write(arg + (char)0);
    bw.write(s, 0, arg.length());
    bw.write((char) 0);
    bw.flush();
    
    // wait for reply
		do {
			if (is.available() > 0) {
        n = is.read(b);  

        // break if connection broken
        if (n == -1 || n == 0) {
          break;
				}
        if (b[n-1] == 0) {
          // break if end of string received  
          reply += new String(b, 0, n-1);
          break;
        } else {
          // add partial reply
          reply += new String(b, 0, n);
				}
      } else {
				try { 
					Thread.sleep(10); 
				} catch (Exception e) {	// need to consider if this locks up
					// do nothing
				}
			}
    } while (true);

    return reply;
  }

	/**
	 * Receives any events for theis client
	 *
	 * @param hBuf
	 * @throws IOException
	 * @return  a byte array of event info */	
  public synchronized byte[] rpcReceiveEvent(int hBuf) throws IOException {

    int i;

    // send rpc command
    pw.print("bm_receive_event&"+hBuf+"&10000&1"+(char)0);
    pw.flush();
    
    // receive reply
    String reply = "";
    do
      {
      i = is.read();  
      if (i == -1 || i == 0)
        break;
          
      reply += (char) i;
      } while (true);
    
    int n = Integer.parseInt(replyParam(reply, 1));
    
    // receive binary event
    byte array[] = new byte[n];

    if (n>0) {
      int received = 0;
      do {
        if (is.available() > 0) {
          i = is.read(array, i, n-received);  

          // break if connection broken
          if (i == -1 || i == 0)
            break;

          received += i;
        }
      } while (received < n);
    }

    // receive remaining parameters
    do {
      i = is.read();
    } while (i != -1 && i != 0);

    return array;
  }

	/**
	 * Parses a message from Midas and returns the requested parameter. All parameters
	 * are separated by & in the message.
	 *
	 * @return a String with the requested parameter
	 * @param reply the reply from the rpc call that must be parsed
	 * @param index the parameter index in the reply string
	 */	
  public String replyParam(String reply, int index) {
		String param = "";
		int i;
		int begin;
		int end;
	
    // find beginning of param
		for (i=begin=0 ; i<index ; i++) {
			while (begin < reply.length() && reply.charAt(begin) != '&') {
				begin++;
			}
			if (reply.charAt(begin) == '&') {
				begin++;
			}
		}
	  
		// find end of param
		end = begin;
		while (end < reply.length() && reply.charAt(end) != '&') {
			end++;
		}
		return reply.substring(begin, end);
	}
	
	
  private char binToAscii(int i) {
    if (i>=38) {
      i = i-38+(int) 'a';
		} else if (i>=12) {
      i = i-12+(int) 'A';
		} else {
      i = i+'.';
		}
    return (char) i;
  }
  
  private String crypt(String buf, String salt) {
    int i, seed;
    StringBuffer enc_pw = new StringBuffer(salt);

    for (i=0; i<8 && i<buf.length() ; i++) {
		  enc_pw.append(buf.charAt(i));
		}
	  for (; i<11; i++) {
		  enc_pw.append((char) 0);
		}
    seed = 123;
	  for (i=2; i<13; i++) {
		  seed = 5*seed + 27 + (int) enc_pw.charAt(i);
      enc_pw.setCharAt(i, binToAscii(seed & 0x3F));
    }
    return enc_pw.toString();
  }
 
  /**
	 * Send the passed message to the experiment.
	 *
	 * @param type specifies the type of the message
	 * @param msg a String containing the message to send to the ODB
	 * @throws IOException thrown from the rpcCall method
	 */  
  public void messageSend(int type, String msg) throws IOException {
    rpcCall("cm_msg&"+type+"&"+client+"&?&?&"+msg);
  }
  
	/**
	 * Send the passed message to the experiment,
	 *
	 * @param type specifies the type of the message
	 * @param source specifies a source of the message. Can be a client
	 * or method location.
	 * @param msg a String containing the message to send to the ODB
	 * @throws IOException thrown from the rpcCall method
	 */	
  public void messageSend(int type, String source, String msg) throws IOException {
    rpcCall("cm_msg&"+type+"&"+client+"&?&"+source+"&"+msg);
  }

	/**
	 * Retrieves all messages for this client
	 *
	 * @throws IOException
	 * @return  a Vector containing all the returned messages
	 */	
  public Vector retrieveMessages() throws IOException {
    Vector msg = new Vector();
    
    String r = rpcCall("cm_msg_retrieve&1000");
    
    // strip status
    r = r.substring(r.indexOf('&')+1, r.length());

    // copy messages into vector
    int ib = 0;
    int ie;
    
    do {
      ie = r.indexOf('\n', ib);
      if (ie > -1) {
        if (r.charAt(ie-1) == '\r') { // CR-LF msg
          msg.addElement(new String(r.substring(ib, ie-1)));
        }
        else { // LF msg
          msg.addElement(new String(r.substring(ib, ie)));
        }
        ib = ie+1;
      }
    } while (ie > -1);
    
    return msg;
  }
  
  public void addMessageListener(MessageListener ml) throws IOException {
    // open message buffer
    String r = rpcCall("bm_open_buffer&SYSMSG&100000");
    hBufMsg = Integer.parseInt(replyParam(r, 1));
    rpcCall("bm_add_event_request&"+hBufMsg+"&-1&-1&2&1&0");
    rpcCall("bm_mark_read_waiting&1");
    
    messageListener = ml;
  }
  
	/**
	 * Test to see if the passed client name is already connected
	 * to the host.
	 *
	 * @param client the name of the client
	 * @param unique true if there should be only one connection from this client
	 * @throws IOException result of any socket error
	 * @return true if the passed client is connected
	 */	
  public boolean isClient(String client, boolean unique) throws IOException {
		String r = rpcCall("cm_exist&" + client + "&" + (unique ? "1" : "0"));
		return replyParam(r, 0).equals(CM_SUCCESS);
  }
  
	/**
	 * Closes the connection to the Midas host.
	 */
	public void disconnect() {
		try {
			if (hDB > 0) {
				
				// Send stop notification
				messageSend(MINFO, "Program " + client + " on host "+
				InetAddress.getLocalHost().getHostName() + " stopped");
				
				// send shutdown message
				rpcCall("id_exit");
				
				// mark experiment disconnected
				hDB = 0;
			
				timerHotLink.cancel();	// shut down hotLink task
				
				// close sockets
				sockSend.close();
				sockRecv.close();
				sockEvent.close();
			}
		} catch (Exception e) {}
	}

	/**
	 * Monitors the receive socket for incoming messages. When detected,
	 * the message is identified as a hotlink callback or a system message.
	 * Monitors the receive socket on a regular basis using a java.util.Timer.
	 **/
	private void hotLink() {
		// Create and start hotlink timer
		timerHotLink = new java.util.Timer(true);
		timerHotLink.scheduleAtFixedRate(new TimerTask() {
			public void run() {
				
				// check receive socket
				try {
					String str = "";
					int i;
					OdbKey key;
					
					i = recvIs.read();
					
					// if we come here, no exception has been raised, therefore
					// we can increase the timeout to receive the full message
					sockRecv.setSoTimeout(1000);
					
					// receive message until '0' is received
					do {
						str += (char) i;
						i = recvIs.read();
					} while (i > 0);
					
					// reset timeout
					sockRecv.setSoTimeout(1);
					
					StringTokenizer st = new StringTokenizer(str, "&");
					String msg = st.nextToken();
					if (msg.equals("MSG_ODB")) {
						st.nextToken();
						int hKey = Integer.parseInt(st.nextToken());
						
						// lookup key
						for (i=0 ; i<vecHotLink.size() ; i++) {
							key = (OdbKey)vecHotLink.elementAt(i);
							if (key.getHkey() == hKey) {
								key.dispatchUpdate();
							}
						}
					} else if (msg.equals("MSG_BM")) {
						String bufName = st.nextToken();
						
						// receive message(s)
						if (bufName.equals("SYSMSG")) {
							do {
								byte array[] = rpcReceiveEvent(hBufMsg);
								
								// break if no more events
								if (array.length == 0)
									break;
								
								// convert array (without header to string)
								String message = new String(array, 16, array.length-16);
								
								// remove padding
								int end = message.indexOf(0);
								message = message.substring(0, end);
								
								if (messageListener != null) {
									messageListener.receiveMessage(message);
								}
							} while (true);
						} else {
							// receive event ? what code goes here ?
						}
					}
				} catch (SocketException se) {
					se.printStackTrace();
					System.out.println("Connection to server lost, stopping program");
					System.exit(1);
				} catch (Exception ex) {
					// do nothing
				}
			}}, 1000, 100);
	}
	
	
	/**
	 * Add a hotlink into the table for monitoring.
	 *
	 * @param key the ODB key of interest
	 */	
  public void addHotLink(OdbKey key) {
    vecHotLink.addElement(key);
  }

	/**
	 * Removes the passed key from the hotlink table.
	 *
	 * @param key the key to be removed
	 */	
  public void removeHotLink(OdbKey key) {
    vecHotLink.removeElement(key);
  }
  
	/**
	 * Returns a String containg the contents of the specified key.
	 *
	 * @param keyName the name of the key
	 * @throws OdbKeyNotFoundException thrown if the key is not found in the ODB
	 * @throws IOException thrown if an error occurrs in rpcCall
	 * @return a String containing the value of the key
	 */	
	public String getOdbValue(String keyName) throws OdbKeyNotFoundException, IOException {
		OdbKey key = new OdbKey(this, keyName);
		return key.toString();
	}

	/** 
	 * Sets the value of the key to the contents of the passed String.
	 *
	 * @param keyName the name of the key
	 * @param value the new value for the key
	 * @throws OdbKeyNotFoundException thrown if the key is not found in the ODB
	 * @throws IOException thrown if an error occurrs in rpcCall
	 */	
	public void setOdbValue(String keyName, String value) throws OdbKeyNotFoundException, IOException {
		OdbKey key = new OdbKey(this, keyName);
		key.setText(value);
	}

	/**
	 * Returns an OdbKey reference for the key located at the
	 * offset from the startPath location in the ODB.
	 *
	 * @param startPath the starting location in the ODB tree. If specified as zero
	 * then the root of the tree is used.
	 * @param index the offset from the starting path
	 * @throws OdbKeyNotFoundException thrown if the startPath is not valid
	 * @throws IOException thrown if an error occurrs in the rpcCall
	 * @return the key of the location or null if the index is too large for the tree
	 */	
  public OdbKey enumOdbKey(String startPath, int index) throws OdbKeyNotFoundException, IOException {
    OdbKey startKey = new OdbKey(this, startPath);
    String r = rpcCall("db_enum_key&"+hDB+"&"+startKey.getHkey()+"&"+index);
    if (replyParam(r, 0).equals(DB_SUCCESS)) {
      int hSubkey = Integer.parseInt(replyParam(r, 1));
      OdbKey subkey = new OdbKey(this, hSubkey);
      return subkey;
    } else {
			return null;
		}
  }

	
	/**
	 * Creates a record in the ODB using the passed String to define each entry in the record.
	 *
	 * The structure of the definition string is shown in this example:
	 *
	 * <code>
	 * 
	 * </code>
	 *
	 * @param key the key to search for. Can conatin a directory
	 * @param record a String defining the entries for the record
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey createRecord(String key, String record) throws IOException, OdbKeyNotFoundException {

		// check if valid path
    if (key.equals("")) {
      throw new IOException("Key path not set");
		}
		
		OdbKey root = null;
		root = new OdbKey(this, key);
			
		
		// parse the record String for the header info and then each line below
/*		
		StringTokenizer st = new StringTokenizer(record);
		String topDir = st.nextToken("[]").trim();
		OdbKey recordKey = new OdbKey(this, root, topDir);
		while (st.hasMoreTokens ) {
			String entry = st.nextToken("\n");
			OdbKey entryKey = new OdbKey(this, recordKey, entry);
			
			
*/		
			
		
/*
 "[Settings]\n\
level1 = INT : 0\n\
level2 = INT : 0";
	*/	
		
		
    //String r = experiment.rpcCall("db_create_record&"+hDB+"&"+rootKey.hKey+"&"+path+"&"+);
    //this.path = experiment.replyParam(r, 1);
    return null;
	}

	
	
	/**
	 * Waits for the Midas experiment to go into the passed run state, and then returns.
	 *
	 * @param state the state to be reached by the run control
	 */
  public void waitForState(int state) throws TooManyListenersException, 
                OdbKeyNotFoundException, IOException, InterruptedException
  {
		if (state < STATE_STOPPED || state > STATE_RUNNING) {
			throw new OdbKeyNotFoundException("Invalid Run state requested");
		}
    OdbKey k = new OdbKey(this, "/Runinfo/State");
    k.hotLink();
    while (Integer.parseInt(k.toString()) != state)
      Thread.sleep(1000);
  }
  
	/**
	 * Tests to see if the current run state of the experiment matches
	 * the passed state value.
	 *
	 *
	 * @param state the state to be tested for.
	 *
	 * @throws TooManyListenersException
	 * @throws OdbKeyNotFoundException the run state key was not found
	 * @throws IOException
	 * @throws InterruptedException
	 *
	 * @return true if the run state matches the passed state else false
	 */	
  public boolean isState(int state) throws TooManyListenersException, 
                OdbKeyNotFoundException, IOException, InterruptedException
  {
		if (state < STATE_STOPPED || state > STATE_RUNNING) {
			throw new OdbKeyNotFoundException("Invalid Run state requested");
		}
    OdbKey k = new OdbKey(this, "/Runinfo/State");
    return Integer.parseInt(k.toString()) == state;
  }
  
	/**
	 * Starts a run with automatic update of the run number
	 *
	 * @return false if there are no errors, else true
	 */
  public boolean start() {
    try {
      String r = rpcCall("cm_execute&" + odbName + " -e " + experiment + " -c 'start -v'&0&0");
      
      if (!replyParam(r, 0).equals(CM_SUCCESS)) {
        return true;
      }
    } catch (IOException e) {
			System.out.println(e); 
		}
		return false;
  }

	/** 
	 * Starts a run with the passed run number
	 *
	 * @param runNumber the number of the run
	 * @return false if there are no errors, else true
	 */
  public boolean start(int runNumber) {
    try {
      if (runNumber == 0) {	// increment run number
        runNumber = getRunNumber();
			}
      String r = rpcCall("cm_execute&" + odbName + " -e " + experiment + " -c 'start " + (runNumber + 1) + " -v'&0&0");
      
      if (!replyParam(r, 0).equals(CM_SUCCESS)) {
        // reset run number
				setRunNumber(runNumber);
        return true;
      }
    } catch (IOException e) {
			System.out.println(e); 
		}
		return false;
  }

	
	/**
	 * Gets the current run number of the Midas experiment.
	 *
	 * @throws IOException thrown if any error communicating with host
	 * @return the current number of the run
	 */	
	public int getRunNumber() throws IOException {
		OdbKey key = new OdbKey(this, "/Runinfo/Run number");
		return Integer.parseInt(key.toString());
	}

	/**
	 * Sets the run number to the passed value.
	 *
	 * @param runNumber the new run number to be used on start
	 * @throws IOException thrown on any error with RPC communication
	 */	
	public void setRunNumber(int runNumber) throws IOException {
		OdbKey key = new OdbKey(this, "/Runinfo/Run number");
    key.setText(key.toString());
	}
		
		
/*	
  public void startDialog(Component parent) {
        
	try {
      // check if transition in progress
      OdbKey trans = new OdbKey(this, "/Runinfo/Transition in progress");
      if (trans.toString().equals("1"))
        JOptionPane.showMessageDialog(parent, 
          "Start/Stop already in progress, please try again later or set\n"+
          "\"/Runinfo/Transition in progress\" manually to zero.",
          "JMIDAS experiment control",
          JOptionPane.WARNING_MESSAGE, null);
      else {
        trans.setText("1");

        JPanel p[] = new JPanel[10];
        JLabel l[] = new JLabel[10];
        MJTextField  t[] = new MJTextField[10];
        
        try {
          // run number field with incremented run number
          p[0] = new JPanel(new GridLayout(1, 2));    
          l[0] = new JLabel("Run number");
          t[0] = new MJTextField("/Runinfo/Run number", false);
          t[0].connect(this);
          int rn = Integer.parseInt(t[0].getData())+1;
          t[0].setData(""+rn);
          p[0].add(l[0]);
          p[0].add(t[0]);

          // loop through experiment parameters
          OdbKey expParam;
          for (int i=0 ; ; i++) {
            expParam = enumOdbKey("/Experiment/Edit on start", i);
            if (expParam == null)
            break;
              
            p[i+1] = new JPanel(new GridLayout(1, 2));
            l[i+1] = new JLabel(expParam.getName()+":");
            t[i+1] = new MJTextField(expParam.getPath(), false);
            t[i+1].connect(this);
            p[i+1].add(l[i+1]);
            p[i+1].add(t[i+1]);
          }

        } catch (OdbKeyNotFoundException e) {};
        
        int reply = JOptionPane.showConfirmDialog(parent, 
                      p, "Enter run parameters" , 
                      JOptionPane.OK_CANCEL_OPTION, 
                      JOptionPane.PLAIN_MESSAGE);
        if (reply == JOptionPane.OK_OPTION) {
          try {
            for (int i=1 ; t[i] != null ; i++)
            t[i].propagate();
          } catch (IOException e) { System.out.println(e); }

          // start run with new run number
          start(Integer.parseInt(t[0].getData()));
        }
        
        trans.setText("0");
	  }
	} catch (Exception e) { System.out.println(e); };
  }

  public void stopDialog(Component parent) {
	try {
      // check if transition in progress
      OdbKey trans = new OdbKey(this, "/Runinfo/Transition in progress");
      if (trans.toString().equals("1"))
        JOptionPane.showMessageDialog(parent, 
          "Start/Stop already in progress, please try again later or set\n"+
          "\"/Runinfo/Transition in progress\" manually to zero.",
          "JMIDAS experiment control",
          JOptionPane.WARNING_MESSAGE, null);
      else {
        trans.setText("1");
	    stop();
        trans.setText("0");
	  }
	} catch (Exception e) { System.out.println(e); };
  }
*/
	/**
	 * Stops the current run in progress
	 *
	 * @return true if an error in the call to Midas else false
	 */
  public boolean stop() {
    try {
      String r = rpcCall("cm_execute& -c 'stop -v'&0&0");
      if (!replyParam(r, 0).equals(CM_SUCCESS)) {
				return true;
      }
    } catch (IOException e) {
			System.out.println(e); 
		}
		return false;
  }

	/**
	 * Pauses the current run in progress
	 *
	 * @return true if an error in the call to Midas else false
	 */
  public boolean pause() {
    try {
      String r = rpcCall("cm_execute&" + odbName + " -e " + experiment + " -c 'pause -v'&0&0");
//      String r = rpcCall("cm_transition&4&0&256&0");
      if (!replyParam(r, 0).equals(CM_SUCCESS)) {
				return true;
      }
    } catch (IOException e) {
			System.out.println(e); 
		}
		return false;
  }

	/**
	 * Resumes a paused run
	 *
	 * @return true if an error in the call to Midas else false
	 */
  public boolean resume() {
    try {
      String r = rpcCall("cm_execute&" + odbName + " -e " + experiment + " -c 'resume -v'&0&0");
//      String r = rpcCall("cm_transition&8&0&256&0");
      if (!replyParam(r, 0).equals(CM_SUCCESS)) {
				return true;
      }
    } catch (IOException e) {
			System.out.println(e); 
		}
		return false;
	}

	/** Ensures that the connection to Midas is closed
	 * @throws Exception thrown from any socket error
	 */
  protected void finalize() throws Exception {
    disconnect();
  }

  /** main method - Used for unit testing
	 * @param args not used
	 */
  public static void main(String args[]) {
   	System.out.println("Midas Experiment Unit Tests starting");
		
		String client = "UnitTest";
		String experiment = "Test";
		String host = "";
		Experiment exp = new Experiment(host, experiment, client);
		try {
			exp.connect();
			System.out.println("Client " + (exp.isClient(client, false) ? "" : "not ") + "connected");
			OdbKey key = new OdbKey(exp, "/Runinfo/Start time");
			System.out.println("Value of " + key.getName() + " is " + key.toString());
			
			OdbKey keyTest = new OdbKey(exp, "/Test", OdbKey.TID_INT);
			keyTest.setText("123");
			System.out.println("Value of " + keyTest.getName() + " is " + keyTest.toString());
			keyTest.delete();
			try {
				keyTest = new OdbKey(exp, "/Test");
			} catch(OdbKeyNotFoundException e) {
				System.out.println("The /Test key was removed by unit test");
			}

			OdbKey keyTest1 = new OdbKey(exp, "/Test1", OdbKey.TID_INT, 5);
			OdbKey keyTest2 = new OdbKey(exp, "/Test2", OdbKey.TID_BOOL, 3);
			OdbKey keyTest3 = new OdbKey(exp, "/Test3", OdbKey.TID_FLOAT, 3);
			OdbKey keyTest4 = new OdbKey(exp, "/Test4", OdbKey.TID_DOUBLE, 4);
//			keyTest1.setTextAt("123",0);
//			keyTest1.setTextAt("234",1);
//			keyTest1.setTextAt("345",2);
//			System.out.println(keyTest1.getPath()+" "+keyTest1.getType()+" "+keyTest1.getNumValues()+" "+keyTest1.toString());
			
			
//			OdbKey keyRecord = new OdbKey(exp, "/Equipment/Table","[X Axis] position = FLOAT : 0.0;pos lim = BOOL : no;neg lim = BOOL : no;dest = FLOAT : 0.0;");
			OdbKey keyTable = new OdbKey(exp, "/Equipment/Table/X Axis", OdbKey.TID_KEY);
			new OdbKey(exp, keyTable, "position", OdbKey.TID_FLOAT);
			new OdbKey(exp, keyTable, "destination", OdbKey.TID_FLOAT);
			new OdbKey(exp, keyTable, "pos limit", OdbKey.TID_BOOL);
			new OdbKey(exp, keyTable, "neg limit", OdbKey.TID_BOOL);


			exp.start(0);
			
			
//			exp.stop();
			
			exp.disconnect();
		} catch (Exception e) {
			System.out.println(e);
		}
   	System.out.println("Midas Experiment Unit Tests completed");
  }
}

