package ca.triumf.midas;

import java.io.*;
import java.util.*;

/**
 * The OdbKey class defines the state and behaviour for a Midas ODB key. 
 * Tools are available for finding and making keys, getting and setting data values,
 * attaching and removing hotlinks for a variable, and examining all details about a key.
 * <P>
 * The following example shows some typical manipulation of keys.
 * <P>
 *<B><code><pre>
public static void main(String args[]) {
  OdbKey keyTest;
  Experiment exp = new Experiment("localhost", "exp123", "JMidas");
  try {
    exp.connect();
    
    keyTest = new OdbKey(exp, "/Test", OdbKey.TID_INT);
    keyTest.setText("123");
    System.out.println("Value of " + keyTest.getName() + " is " + keyTest.toString());
    keyTest.delete();
    keyTest = null;
    keyTest = new OdbKey(exp, "/Test");

    exp.disconnect();
  } catch (OdbKeyNotFoundException e) {
    System.out.println("Did not find " + e);
  } catch (Exception e) {
    System.out.println(e);
  }
}
 * </pre></code></B>
 * <P>
 * The following example shows typical manipulation of a key array.
 * <P>
 *<B><code><pre>
public static void main(String args[]) {
  OdbKey keyTest;
  Experiment exp = new Experiment("localhost", "exp123", "JMidas");
  try {
    exp.connect();
    keyTest = new OdbKey(exp, "/Test", OdbKey.TID_INT, 5);
    keyTest.getInfo();
    System.out.println(keyTest.getPath() + " has " + keyTest.getNumValues() + " values in the array");
    exp.disconnect();
  } catch (Exception e) {
    System.out.println(e);
  }
}
 * </pre></code></B>
 *
 * @author Stefan Ritt - PSI
 * @author David Morris - TRIUMF
 */
public class OdbKey extends java.lang.Object {
  
  public static final int TID_BYTE    =  1; /* unsigned byte         0       255    */
  public static final int TID_SBYTE   =  2; /* signed byte         -128      127    */
  public static final int TID_CHAR    =  3; /* single character      0       255    */
  public static final int TID_WORD    =  4; /* two bytes             0      65535   */
  public static final int TID_SHORT   =  5; /* signed word        -32768    32767   */
  public static final int TID_DWORD   =  6; /* four bytes            0      2^32-1  */
  public static final int TID_INT     =  7; /* signed dword        -2^31    2^31-1  */
  public static final int TID_BOOL    =  8; /* four bytes bool       0        1     */
  public static final int TID_FLOAT   =  9; /* 4 Byte float format                  */
  public static final int TID_DOUBLE  = 10; /* 8 Byte float format                  */
  public static final int TID_BITFIELD= 11; /* 32 Bits Bitfield      0  111... (32) */
  public static final int TID_STRING  = 12; /* zero terminated string               */
  public static final int TID_ARRAY   = 13; /* array with unknown contents          */
  public static final int TID_STRUCT  = 14; /* structure with fixed length          */
  public static final int TID_KEY     = 15; /* key in online database               */
  public static final int TID_LINK    = 16; /* link in online database              */
  public static final int TID_LAST    = 17; /* end of TID list indicator            */

  private int hDB;
  private int hKey;
  private boolean valid;
  private boolean dataValid;
  private boolean hotLinked;
  private String path;
  private String name;
  private int type;
  private int numValues;
  private int itemSize;
  private String data;
  private String array[];
  
  private Experiment experiment;
  private OdbKeyUpdate keyUpdate;
  
	/**
	 * Default constructor
	 */
  public OdbKey() {
    valid = false;
  }
  
	/** 
	 * Creates a handle to an existing ODB key
	 *
	 * @param experiment the experiment where the key is located
	 * @param rootKey the root key that the path starts at
	 * @param path the path to the key with the key name as the last item in the path
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, OdbKey rootKey, String path) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = rootKey.path + "/" + path;
    
    String r = experiment.rpcCall("db_find_key&"+hDB+"&"+rootKey.hKey+"&"+path);
    if (experiment.replyParam(r, 0).equals("1")) {
      valid = true;
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+rootKey+"/"+path);
		}
  }

	/**
	 * Creates a handle to an existing ODB key
	 *
	 * @param experiment the experiment where the key is located
	 * @param path the fully qualified path to the key with the key name as the last item in the path 
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, String path) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = path;
    
    String r = experiment.rpcCall("db_find_key&"+hDB+"&0&"+path);
    if (experiment.replyParam(r, 0).equals("1")) {
      valid = true;
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
      
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+path);
		}
  }

	/** 
	 * Creates a handle to an ODB key and creates the key if it does not exist.
	 *
	 * @param experiment the experiment where the key is located
	 * @param path the fully qualified path to the key with the key name as the last item in the path
	 * @param type the type of the key. See the TID declarations for this class.
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, String path, int type) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
		if ((type <= 0) || (type > TID_LAST)) {	// check for valid key type
      throw new IOException("Invalid key type");
		}
			
    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = path;

    // create key with given type
    experiment.rpcCall("db_create_key&"+hDB+"&0&"+path+"&"+type);

    String r = experiment.rpcCall("db_find_key&"+hDB+"&0&"+path);
    if (experiment.replyParam(r, 0).equals("1")) {
      valid = true;
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+path);
		}
  }

	/**
	 * Creates a handle to an ODB key and creates the key if it does not exist.
	 *
	 * @param experiment the experiment where the key is located
	 * @param rootKey the root key that the path starts at
	 * @param path the fully qualified path to the key with the key name as the last item in the path
	 * @param type the type of the key. See the TID declarations for this class.
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, OdbKey rootKey, String path, int type) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
		if ((type <= 0) || (type > TID_LAST)) {	// check for valid key type
      throw new IOException("Invalid key type");
		}

    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = path;

		String fullPath = rootKey.getPath() + "/" + path;
		
    // create key with given type
    experiment.rpcCall("db_create_key&"+hDB+"&0&"+fullPath+"&"+type);

    String r = experiment.rpcCall("db_find_key&"+hDB+"&0&"+fullPath);
    if (experiment.replyParam(r, 0).equals("1")) {
      valid = true;
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+fullPath);
		}
  }

	/**
	 * Creates a handle to an ODB array and creates the array if it does not exist.
	 *
	 * @param size the size of the array
	 * @param experiment the experiment where the key is located
	 * @param path the fully qualified path to the key with the key name as the last item in the path
	 * @param type the type of the key. See the TID declarations for this class.
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, String path, int type, int size) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
		if ((type <= 0) || (type > TID_LAST)) {	// check for valid key type
      throw new IOException("Invalid key type");
		}
		if ((size <= 0) || (size > 10000)) {	// check for crazy array size
      throw new IOException("Array size too big");
		}
		
    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = path;
    
    // create key with given type
    experiment.rpcCall("db_create_key&"+hDB+"&0&"+path+"&"+type);

    String r = experiment.rpcCall("db_find_key&"+hDB+"&0&"+path);
    if (experiment.replyParam(r, 0).equals("1")) {
      valid = true;
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
	    experiment.rpcCall("db_set_data&"+hDB+"&"+hKey+"&0&0&"+size+"&"+type);
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+path);
		}
  }

	/** 
	 * Creates an ODB record from the passed definition string.
	 *
	 * @param experiment the experiment where the key is located
	 * @param path the fully qualified path to the key with the key name as the last item in the path
	 * @param type the type of the key. See the TID declarations for this class.
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, String path, String record) throws IOException, OdbKeyNotFoundException {
    if (path.equals("")) {	// check if valid path
      throw new IOException("Key path not set");
		}
    this.experiment = experiment;
    hDB = experiment.getHdb();
    valid = false;
    dataValid = false;
    hotLinked = false;
    this.path = path;
    
    // create key with given type
    experiment.rpcCall("db_create_record&"+hDB+"&0&"+path+"&"+record);

    String r = experiment.rpcCall("db_find_key&"+hDB+"&0&"+path);
    if (experiment.replyParam(r, 0).equals("1")) {
      hKey = Integer.parseInt(experiment.replyParam(r, 1));
      getInfo();
    } else {
			throw new OdbKeyNotFoundException("OdbKey "+path);
		}
  }

	/**
	 * Creates a handle to a key from an existing key handle
	 *
	 * @param experiment the experiment where the key is located
	 * @param hKey a handle to an existing key
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  public OdbKey(Experiment experiment, int hKey) throws IOException, OdbKeyNotFoundException {
    if (hKey == 0) {	// check if valid path
      throw new OdbKeyNotFoundException("Invalid hKey=0");
		}
    this.experiment = experiment;
    this.hDB = experiment.getHdb();
    this.hKey = hKey;

    // retrieve full path
    String r = experiment.rpcCall("db_get_path&"+hDB+"&"+hKey+"&256");
    this.path = experiment.replyParam(r, 1);
    
    valid = true;
    dataValid = false;
    hotLinked = false;
      
    getInfo();
  }

	/**
	 * Creates an ODB key from an existing key handle. An optimized version used 
	 * by enumSubKey to avoid retrieval of path.
	 *
	 * @param experiment the experiment where the key is located
	 * @param hKey a handle to an existing key
	 * @param parent the parent key of the handle
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbKeyNotFoundException thrown if the root or path is invalid
	 */	
  private OdbKey(Experiment experiment, int hKey, OdbKey parent) throws 
            IOException, OdbKeyNotFoundException {

    // check if valid path
    if (hKey == 0)
      throw new OdbKeyNotFoundException("Invalid hKey=0");

    this.experiment = experiment;
    this.hDB = experiment.getHdb();
    this.hKey = hKey;
    valid = true;
    dataValid = false;
    hotLinked = false;
    
    getInfo();

    // construct full path
    this.path = parent.getName() + "/" + name;
  }
	
	/**
	 * Returns the handle of the key
	 *
	 * @return the handle of the key
	 */	
  public int getHkey() {
    return hKey;
  }
  
  /**
   * Returns the full path name of the key. 
   *
   * @return the full path name of the key
   */
  public String getPath() {
    return path;
  }

	/**
	 * Returns the type of the key. 
	 * The int value must be matched to the TID_xxxx Midas data type.
	 *
	 * @return the type of the key as an int value
	 */
  public int getType() {
    return type;
  }

	/**
	 * Returns the number of values in a key. Intended to query the size of an array
	 * for access of manipulation of the data
	 *
	 * @return the number of values in the array
	 */
  public int getNumValues() {
    return numValues;
  }

	/**
	 * Returns the name of the key. 
	 *
	 * @return the name of the key, excluding the prefixed path.
	 */
  public String getName() {
    return name;
  }

	/**
	 * Updates the information for this key. Updates the name, type, number of values and item size.
	 *
	 * @throws IOException thrown by socket errors
	 */
  public void getInfo() throws IOException {
    String r = experiment.rpcCall("db_get_key_info&"+hDB+"&"+hKey+"&32");
    if (experiment.replyParam(r, 0).equals("1")) {
      this.name = experiment.replyParam(r, 1);
      this.type = Integer.parseInt(experiment.replyParam(r, 2));
      this.numValues = Integer.parseInt(experiment.replyParam(r, 3));
      this.itemSize = Integer.parseInt(experiment.replyParam(r, 4));
    }
  }

	/**
	 * Updates the data associated with this key
	 *
	 * @throws IOException thrown by socket errors
	 */
  public void getData() throws IOException {
    if (type == TID_KEY) // key doesn't contain data
      data = null;
    else {
      String r = experiment.rpcCall("db_get_data1&"+hDB+"&"+hKey+"&64000&"+type);
      if (experiment.replyParam(r, 0).equals("1")) {
        this.numValues = Integer.parseInt(experiment.replyParam(r, 1));
        if (this.numValues == 1)
          this.data = experiment.replyParam(r, 2);
        else {
          this.array = new String[numValues];        
          for (int i=0 ; i<numValues ; i++)
            this.array[i] = experiment.replyParam(r, 2+i);
        }
      }
    }
    dataValid = true;
  }

	/**
	 * Updates the data for the index entry of this key
	 *
	 * @throws IOException thrown by socket errors
	 */
  public void getDataAt(int index) throws IOException {
    
    if (type == TID_KEY) // key doesn't contain data
      data = null;
    else {
      String r = experiment.rpcCall("db_get_data_index&"+hDB+"&"+hKey+"&64000&"+index+"&"+type);
      if (experiment.replyParam(r, 0).equals("1")) {
        if (numValues == 1)
          this.data = experiment.replyParam(r, 2);
        else
          this.array[index] = experiment.replyParam(r, 2);
      }
    }
  }

	/**
	 * Sets the value of the variable to the passed string. Care must be taken
	 * that the passed value makes sense in terms of the type of the ODB key.
	 *
	 * @param str a String describing the new value
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbNoAccessException thrown if key is write-locked
	 */
  public void setText(String str) throws IOException, OdbNoAccessException {
    if (numValues > 1) {

      // break string apart into individual values
      int i, n;
      
      StringTokenizer st = new StringTokenizer(str, "[], ");
      n = st.countTokens();
      String tmp[] = new String[n];
      
      String rpc = "db_set_data&"+hDB+"&"+hKey+"&"+n+" "+type;
      
      for (i=0 ; i<n ; i++) {
        tmp[i] = st.nextToken();
        rpc += "&"+tmp[i];
      }
      rpc += "&0&"+n+"&"+type;
      String r = experiment.rpcCall(rpc);
      if (experiment.replyParam(r, 0).equals("318")) {
        throw new OdbNoAccessException("OdbKey "+this.path);
			}
			
      // if successful, copy data to this  
      array = new String[n];  
      for (i=0 ; i<n ; i++) {
        array[i] = tmp[i];
			}
    } else {
      
      // send individual value
      String r = experiment.rpcCall("db_set_data&"+hDB+"&"+hKey+
                                    "&1 "+type+"&"+str+"&0&1&"+type);
      if (experiment.replyParam(r, 0).equals("318")) {
        throw new OdbNoAccessException("OdbKey "+this.path);
			}
      this.data = str;  
    }
  }
  
	/**
	 * Sets the value of the parameter to the passed string. Care must be taken
	 * that the passed value makes sense in terms of the type of the ODB key.
	 *
	 * @param str a String describing the new value
	 * @param index the index of the array entry
	 *
	 * @throws ArrayIndexOutOfBoundsException thrown if the index value is illegal
	 * @throws IOException thrown by socket errors
	 * @throws OdbNoAccessException thrown if key is write-locked
	 */
  public void setTextAt(String str, int index) throws 
                IOException,ArrayIndexOutOfBoundsException,
                OdbNoAccessException {
    if (numValues == 1)
      setText(str);
    else {
      if (index < numValues) {
        String r = experiment.rpcCall("db_set_data_index&"+hDB+"&"+hKey+
                                      "&1 "+type+"&"+str+"&0&"+index+"&"+type);
        if (experiment.replyParam(r, 0).equals("318")) {
          throw new OdbNoAccessException("OdbKey "+this.path);
				}
        this.array[index] = str;
      } else {
				throw new IndexOutOfBoundsException("OdbKey " + this.path);
			}
    }
  }

	/**
	 * Sets the value of the parameter to the passed string. Care must be taken
	 * that the passed value makes sense in terms of the type of the ODB key.
	 *
	 * @param str a String describing the new value
	 * @param index the index of the array entry
	 * @param size the size to grow the array to if not big enough
	 *
	 * @throws ArrayIndexOutOfBoundsException thrown if the index value is illegal
	 * @throws IOException thrown by socket errors
	 * @throws OdbNoAccessException thrown if key is write-locked
	 */
  public void setTextAt(String str, int index, int size) throws 
                IOException,ArrayIndexOutOfBoundsException,
                OdbNoAccessException {

    String r = experiment.rpcCall("db_set_data_index&"+hDB+"&"+hKey+
                                  "&1 "+type+"&"+str+"&"+size+"&"+index+"&"+type);
    if (experiment.replyParam(r, 0).equals("318")) {
      throw new OdbNoAccessException("OdbKey "+this.path);
		}
    getData();
  }

	/**
	 * Sets the value of the parameter to the passed string. Care must be taken
	 * that the passed value makes sense in terms of the type of the ODB key.
	 *
	 * @param str a String describing the new value
	 * @param index the index of the array entry
	 * @param notify if true then all clients are notified of the value change
	 *
	 * @throws ArrayIndexOutOfBoundsException thrown if the index value is illegal
	 * @throws IOException thrown by socket errors
	 * @throws OdbNoAccessException thrown if key is write-locked
	 */
  public void setTextAt(String str, int index, boolean notify) throws 
                IOException,ArrayIndexOutOfBoundsException,
                OdbNoAccessException {
    if (index < numValues) {
      String r = experiment.rpcCall("db_set_data_index2&"+hDB+"&"+hKey+
                                    "&1 "+type+"&"+str+"&0&"+index+"&"+type+(notify?"&1":"&0"));
      if (experiment.replyParam(r, 0).equals("318"))
        throw new OdbNoAccessException("OdbKey "+this.path);
      array[index] = str;
    } else throw new IndexOutOfBoundsException("OdbKey " + this);
  }

	/**
	 * Change the size of an array to the specified number of values
	 *
	 * @param numValues the new number of elements in the array
	 *
	 * @throws IOException thrown by socket errors
	 * @throws OdbNoAccessException thrown if key is write-locked
	 */
  public void setNumValues(int numValues) throws IOException, OdbNoAccessException {
    String r = experiment.rpcCall("db_set_num_values&"+hDB+"&"+hKey+"&"+numValues);
    if (experiment.replyParam(r, 0).equals("318"))
      throw new OdbNoAccessException("OdbKey "+ this.path);
    getData();
  }
  
	/**
	 * Attaches a hotlink to this key. The passed OdbKeyUpdate reference will be attached
	 * to the hotlink monitoring thread. If the value of the key changes, a message is sent to the client
	 * and the update method of the reference is called.
	 *
	 * @param ku an OdbKeyUpdate class reference with an <code>update</code> method
	 *
	 * @throws IOException thrown by socket errors
	 * @throws TooManyListenersException thrown if the ODB has too many registered listeners
	 */
  public void hotLink(OdbKeyUpdate ku) throws IOException, TooManyListenersException {

    // get inital data
    if (!dataValid)
      getData();
      
    String r = experiment.rpcCall("db_add_open_record&"+hDB+"&"+hKey+"&0");
    if (!experiment.replyParam(r, 0).equals("1"))
      throw new TooManyListenersException("OdbKey "+path);
      
    // register hotlink in experiment
    experiment.addHotLink(this);
    
    // save update function
    keyUpdate = ku;
    
    hotLinked = true;
  }

	/**
	 * Attaches a hotlink to this key. The value in the client application will be updated
	 * whenever the ODB value changes.
	 *
	 * @throws IOException thrown by socket errors
	 * @throws TooManyListenersException thrown if the ODB has too many registered listeners
	 */
  public void hotLink() throws IOException, TooManyListenersException {
    hotLink(null);
  }

	/**
	 * Detaches a hotlink for this key.
	 *
	 * @throws IOException thrown by socket errors
	 */
  public void hotUnlink() throws IOException {
    if (hotLinked) {
      experiment.removeHotLink(this);
      experiment.rpcCall("db_remove_open_record&"+hDB+"&"+hKey);
      
      hotLinked = false;
    }
  }

	/**
	 * Updates the data for this key and triggers the update task if it exists.
	 *
	 * @throws IOException thrown by socket errors
	 */
  public void dispatchUpdate() throws IOException {
    getData();
    if (keyUpdate != null)
      keyUpdate.update(this);
  }
  
	/**
	 * Returns the key information in a String. If the key is a directory entry, the name
	 * is returned. If the key is a data item, then the current data value is returned.
	 *
	 * @return a String containing the name for a directory entry or the current value for a data item. If the
	 * data item has more than one value, the data is comma separated.
	 */	
  public String toString() {
    if (!dataValid) {	// if data not yet fetched, do now
      try {
        getData();
      } catch (IOException e) {
				System.out.println(e); 
			}
		}
    if (type == TID_KEY) {	// return name instead data for keys
      return name;
		}
      
    if (numValues == 1) {	// return single value  
      return data;
		} else {	// return comma separated values
      String str = "[";
      int i;
      
      for (i=0 ; i<numValues ; i++) {
        str += array[i];
        if (i<numValues-1) {
          str += ',';
				}
      }
      str += "]";
      return str;
    }
  }

	/**
	 * Returns the key information in a String. If the key is a directory entry, the name
	 * is returned. If the key is a data item, then the current data value for the passed index
	 * is returned.
	 *
	 * @param index the index into an array of data items
	 * @return a String containing the name for a directory entry or the current value for the index of the data item.
	 */	
  public String toString(int index) {
    if (!dataValid) {	// if data not yet fetched, do now
      try {
        getData();
      } catch (IOException e) {
				System.out.println(e);
			}
		}
    if (type == TID_KEY) {	// return name instead data for keys
      return name;
		}
    if (numValues == 1) {	// return single value  
      return data;
		} else {
      return array[index];
		}
  }

  public OdbKey enumSubKey(int index) throws OdbKeyNotFoundException, IOException {
    String r = experiment.rpcCall("db_enum_key&"+hDB+"&"+this.hKey+"&"+index);
    if (experiment.replyParam(r, 0).equals("1")) {
      int hSubkey = Integer.parseInt(experiment.replyParam(r, 1));
      OdbKey subkey = new OdbKey(experiment, hSubkey, this);
      return subkey;
    } else {
			return null;
		}
  }
	
	/**
	 * Removes the key from the database and removes any hotlink associated with the key.
	 *
	 * @throws IOException thrown on any RPC error
	 */	
  public void delete() throws IOException {
    experiment.removeHotLink(this);
    experiment.rpcCall("db_delete_key&"+hDB+"&"+this.hKey+"&0");
  }
  
	/**
	 * Removed any hot link associated with this key
	 *
	 * @throws Exception any RPC errors generated by experiment class
	 */	
  protected void finalize() throws Exception {
    experiment.removeHotLink(this);
  }
}
